# Copyright (c) 2016, Patrick Uiterwijk <patrick@puiterwijk.org>
# All rights reserved.
#
# This file is part of Basset.
#
# Basset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Basset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Basset.  If not, see <http://www.gnu.org/licenses/>.

import logging

from basset.decision.base import DecisionPlugin


class MediawikiDecision(DecisionPlugin):
    name = 'mediawiki'
    handled_decisions = ['mediawiki.edit', 'mediawiki.new']

    def tests_to_run(self, action, time, data):
        return ['content', 'ip']

    def generate_decisions(self, action, time, data, decision):
        if decision is None:
            return None
        elif decision is True:
            return {'action': action,
                    'page': data['page']['url'],
                    'revision': data.get('revision', None),
                    'decision': 'accept'}
        else:
	    return [{'action': 'mediawiki.block',
		     'username': data['page']['user'].lower(),
		     'decision': 'deny'},
		    {'action': action,
		     'page': data['page']['title'],
		     'revision': data.get('revision', None),
		     'decision': 'deny'},
		    {'action': 'fedora.fas.registration',
		     'username': data['page']['user'].lower(),
		     'decision': 'deny'}]

    def _get_csrf_token(self, page):
        self.log.debug('Getting CSRF token for deleting of %s' % page)

        ver = self.core.wikiversion
        if ver[1] <= 24:
            self.log.debug('Old version')
            info = self.core.wikiclient.raw_api('query', titles=page, prop='info', intoken='delete')
            info = info['query']['pages']
            info = info[info.keys()[0]]
            token = info['deletetoken']
        else:
            # The author of Basset does not have access to a new version...
            # The old code does still work for now, so you could bump up the ver[1] check above
            raise Exception('This code is untested')
            self.log.debug('New version')
            info = self.core.wikiclient.raw_api('query', meta='tokens', type='csrf')
            token = info['query']['tokens']['csrftoken']

        self.log.debug('Token: %s' % token)

        return token

    def _delete_page(self, page):
        token = self._get_csrf_token(page)
        result = self.core.wikiclient.raw_api('delete',
                              title=page,
                              token=token,
                              reason='Automatically deleted by Basset')
        if 'error' in result and result['error']['info'] != "The page you requested doesn't exist":
            self.log.error('Error deleting: %s' % result['error']['info'])
            return False
        else:
            return True

    def _revert_page(self, page):
        raise NotImplementedError()

    def handle(self, decision):
        if not self.coreconfig.get('core', 'do_wiki') == 'true':
            self.log.info('Ignoring wiki decisions')
            return True

        if decision['decision'] == 'deny':
            if decision['action'] == 'mediawiki.new':
                return self._delete_page(decision['page'])
            elif decision['action'] == 'mediawiki.edit':
                return self._delete_page(decision['page'])
                #return self._revert_page(decision['page'], decision['revision'])
        elif decision['decision'] == 'manual':
            # TODO: Send notification to the admin
            self.error('Mediawiki page requires manual check!')
            raise NotImplementedError()
        elif decision['decision'] == 'accept':
            return True  # Since mediawiki does stuff before notifying, we do nothing here
        else:
            raise ValueError('Invalid decision value: "%s"' % decision['decision'])

class MediawikiBlockDecision(DecisionPlugin):
    name = 'mediawikiblock'
    handled_decisions = ['mediawiki.block']

    def _get_block_token(self, user):
        # Getting block tokens has had three different APIs over various mediawiki
        # versions.... Let the fun begin!
        self.log.debug('Getting block token for %s' % user)

        ver = self.core.wikiversion
        if ver[1] < 20:
            self.log.debug('Ancient version')
            info = self.core.wikiclient.raw_api('block', user=user, gettoken='1')
            blocktoken = info['block']['blocktoken']
        elif ver[1] >= 20 and ver[1] < 24:
            # The author of Basset has never been able to test with this version.
            # When you encounter this, please report and we can test whether it works.
            raise Exception('Need testing with this version!')
            self.log.debug('Old version')
            info = self.core.wikiclient.raw_api('tokens', type='block')
            blocktoken = info['tokens']['blocktoken']
        elif ver[1] >= 24:
            self.log.debug('New version')
            info = self.core.wikiclient.raw_api('query', meta='tokens', type='csrf')
            blocktoken = info['query']['tokens']['csrftoken']

        self.log.debug('Blocktoken: %s' % blocktoken)
        return blocktoken

    def _block_user(self, user):
        blocktoken = self._get_block_token(user)

        result = self.core.wikiclient.raw_api('block',
                              user=user,
                              token=blocktoken,
                              reason='Automatically blocked by Basset')
        self.log.debug('Block result: %s' % result)
        if 'error' in result and result['error']['code'] != 'alreadyblocked':
            self.log.error('Error blocking: %s' % result['error']['info'])
            return False
        else:
            return True

    def handle(self, decision):
        if not self.coreconfig.get('core', 'do_wiki') == 'true':
            self.log.info('Ignoring wiki decisions')
            return True

        if decision['decision'] == 'deny':
            return self._block_user(decision['username'])
        else:
            self.log.debug('Nothing to do for action "%s"' % decision['decision'])
            return True
