# Copyright (c) 2016, Patrick Uiterwijk <patrick@puiterwijk.org>
# All rights reserved.
#
# This file is part of Basset.
#
# Basset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Basset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Basset.  If not, see <http://www.gnu.org/licenses/>.

import logging
import requests
import datetime
import time

from basset.decision.base import DecisionPlugin


class TracDecision(DecisionPlugin):
    name = 'trac'
    handled_decisions = ['trac.ticket.new', 'trac.ticket.update',
                         'trac.ticket.delete', 'trac.wiki.page.new',
                         'trac.wiki.page.update', 'trac.wiki.page.delete',
                         'trac.wiki.page.rename', 'trac.wiki.page.version.delete',
                         'trac.attachment.added', 'trac.attachment.deleted',
                         'trac.attachment.reparented', 'trac.blockUser',
                         'trac.blockUser']

    def tests_to_run(self, action, time, data):
        return ['content']

    def generate_decisions(self, action, time, data, decision):
        if action.endswith('.delete') or action.endswith('deleted'):
            return []

        if decision is None:
            return None
        elif decision is True:
            return {'action': action,
                    'username': data['agent'],
                    'instance': data['instance'],
                    'decision': 'accept'}
        else:
           return [{'action': action,
		    'page': data.get('page', None),
		    'ticket': data.get('ticket', None),
		    'attachment': data.get('attachment', None),
		    'instance': data['instance'],
		    'decision': 'deny'},
		   {'action': 'trac.blockUser',
		    'username': data['agent'],
		    'instance': data['instance'],
		    'decision': 'deny'},
		   {'action': 'fedora.fas.registration',
		    'username': data['agent'],
		    'decision': 'deny'}]

    def _call(self, trac_base_url, method, *params):
        self.log.debug('Performing trac cal: %s, params: %s' % (method, params))
        if trac_base_url.endswith('/'):
            trac_base_url = trac_base_url[:-1]
        rpc_url = '%s/login/jsonrpc' % trac_base_url
        idnr = 1235
        req = {'method': method,
               'params': params,
               'id': idnr}
        resp = requests.post(rpc_url, json=req,
                             auth=self.core.tracclient['auth'])
        resp = resp.json()
        self.log.debug('Result" %s' % resp)
        if resp['error'] is not None and resp['error']['code'] != 404:
            self.log.error('Error during call: %s' % resp['error'])
            raise Exception('Call error')
        if resp['id'] != idnr:
            self.log.error('Invalid response ID gotten: %s' % resp['id'])
            raise Exception('Invalid response ID')
        return resp['result']

    def _delete_wiki_page(self, trac_base_url, page, version=None):
        if version:
            version = int(version)
        return self._call(trac_base_url, 'wiki.deletePage', page, version)

    def _delete_ticket(self, trac_base_url, ticket):
        self._call(trac_base_url, 'ticket.delete', ticket)
        return True

    def _delete_ticket_change(self, trac_base_url, ticket, changetime):
        # This means that if the timestamp is not UTC, we crash.
        changetime = changetime.replace('+00:00', '')
        date = datetime.datetime.strptime(changetime, '%Y-%m-%d %H:%M:%S.%f')
        ts = str(time.mktime(date.timetuple()))[:-2] + str(date.microsecond)
        return self._call(trac_base_url, 'basset.deleteTicketChange', ticket, ts)

    def _delete_attachment(self, trac_base_url, parent_realm, parent_id, filename):
        if parent_realm == 'ticket':
            return self._call(trac_base_url, 'ticket.deleteAttachment', parent_id, filename)
        elif parent_realm == 'wiki':
            return self._call(trac_base_url, 'wiki.deleteAttachment', '%s/%s' % (parent_id, filename))
        else:
            raise NotImplementedError('Unknown realm: %s' % parent_realm)

    def _block_user(self, trac_base_url, username):
        return self._call(trac_base_url, 'basset.blockUser', username)

    def handle(self, decision):
        if not self.coreconfig.get('core', 'do_trac') == 'true':
            self.log.info('Ignoring trac decisions')
            return True

        base_url = decision['instance']['base_url']
        if decision['decision'] == 'deny':
            if decision['action'] == 'trac.blockUser':
                return self._block_user(base_url, decision['username'])
            elif decision['action'] == 'trac.ticket.new':
                return self._delete_ticket(base_url, decision['ticket']['id'])
            elif decision['action'] == 'trac.ticket.update':
                # This assumes that the changetime of the ticket is the time of this
                # particular change.... sounds reasonable to me.
                return self._delete_ticket_change(base_url,
                                                  decision['ticket']['id'],
                                                  decision['ticket']['changetime'])
            elif decision['action'] == 'trac.wiki.page.new':
                return self._delete_wiki_page(base_url,
                                              decision['page']['name'])
            elif decision['action'] == 'trac.wiki.page.update':
                return self._delete_wiki_page(base_url,
                                              decision['page']['name'],
                                              decision['page']['version'])
            elif decision['action'] == 'trac.attachment.added':
                return self._delete_attachment(base_url,
                                               decision['attachment']['parent_realm'],
                                               decision['attachment']['parent_id'],
                                               decision['attachment']['filename'])

            elif decision['action'] in ['trac.ticket.delete',
                                        'trac.wiki.page.delete',
                                        'trac.wiki.page.version.delete',
                                        'trac.attachment.deleted']:
                # These are deletions, so we don't care
                return True
            else:
                log.error('Unknown action: %s' % decision['action'])
                raise NotImplementedError()
        elif decision['decision'] == 'manual':
            # TODO: Send notification to the admin
            self.error('Trac page requires manual check!')
            raise NotImplementedError()
        elif decision['decision'] == 'accept':
            return True  # Since trac does stuff before notifying, we do nothing here
        else:
            raise ValueError('Invalid decision value: "%s"' % decision['decision'])
