# Copyright (c) 2016, Patrick Uiterwijk <patrick@puiterwijk.org>
# All rights reserved.
#
# This file is part of Basset.
#
# Basset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Basset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Basset.  If not, see <http://www.gnu.org/licenses/>.

import requests
import SubnetTree

from basset.score.base import ScorePlugin


class IPScore(ScorePlugin):
    name = 'ip'

    def _do_init(self):
        # Initialize ASN lookup table
        self.subnetTree = None
        if self.config.get('filename_asn') is not None:
            self.subnetTree = SubnetTree.SubnetTree()
            with open(self.config['filename_asn'], 'r') as asnfile:
                for line in asnfile.readlines():
                    line = line.replace('\n', '')
                    split = line.split(' ')
                    if len(split) != 2:
                        continue
                    cidr, asn = split
                    self.subnetTree[cidr] = asn

    def get_score(self, action, time, data):
        scores = {}

        if 'request_headers' in data:
            ip = data['request_headers']['x-forwarded-for']
            if ',' in ip:
                ip = ip.split(',')[0]

            ipinfo = self._get_rdap_data(ip)
            self.log.debug('IPInfo for %s: %s' % (ip, ipinfo))

            if self.subnetTree is not None:
                scores['ip_asn'] = self.get_ip_asn_score(ip)
            scores['ip_country'] = self.get_ip_country_score(ipinfo)
            scores['ip_owner'] = self.get_ip_owner_score(ipinfo)

        return scores

    def _get_rdap_data(self, ip):
        url = 'http://rdap.apnic.net/ip/%s' % ip
        return requests.get(url).json()

    def get_ip_asn_score(self, ip):
        for option in self.config:
            if option.startswith('asn_'):
                mapping[option[len('asn_'):]] = int(self.config[option])
        if not ip in self.subnetTree:
            # This is an unknown subnet.... Which should NEVER Happen
            return 2
        asn = self.subnetTree[ip]
        return mapping.get(asn, 0)

    def get_ip_country_score(self, ipinfo):
        if not 'country' in ipinfo:
            self.log.error('No country in ipinfo?')
            return 0
        mapping = {}
        for option in self.config:
            if option.startswith('ip_country_'):
                mapping[option[len('ip_country_'):]] = self.config[option]
        if ipinfo['country'] in mapping:
            return int(mapping[ipinfo['country']])
        else:
            return 0

    def get_ip_owner_score(self, ipinfo):
        mapping = {}
        for option in self.config:
            if option.startswith('ip_owner_'):
                mapping[option[len('ip_owner_'):]] = self.config[option]
        for owner in mapping:
            if owner in ipinfo.get('name', '').lower():
                return int(mapping[owner])
        else:
            return 0
